Attribute VB_Name = "Module_MemoryMappedFiles"
Option Explicit

Private Declare Function CreateFileMapping Lib "kernel32.dll" Alias "CreateFileMappingA" (ByVal hFile As Long, _
                                         ByVal lpAttributes As Long, ByVal flProtect As Long, _
                                         ByVal dwMaximumSizeHigh As Long, ByVal dwMaximumSizeLow As Long, _
                                         ByVal lpName As String) As Long

Private Declare Function MapViewOfFile Lib "kernel32.dll" (ByVal hFile As Long, ByVal dwDesiredAccess As Long, _
                                         ByVal dwFileOffsetHigh As Long, ByVal dwFileOffsetLow As Long, _
                                         ByVal dwNumberOfBytesToMap As Long) As Long
                     
Private Declare Function UnmapViewOfFile Lib "kernel32.dll" (ByVal lpBaseAddress As Long) As Boolean

Private Declare Function CloseHandle Lib "kernel32" (ByVal hObject As Long) As Long

' Src -> *(Dst)
Private Declare Sub CopyMemoryWrite Lib "kernel32" Alias "RtlMoveMemory" (ByVal Dst As Long, _
                                                                          pSrc As Any, _
                                                                          ByVal ByteLen As Long)
' *(Src) -> Dst
Private Declare Sub CopyMemoryRead Lib "kernel32" Alias "RtlMoveMemory" (pDst As Any, _
                                                                         ByVal Src As Long, _
                                                                         ByVal ByteLen As Long)

Public Declare Sub CopyMemory Lib "kernel32" Alias "RtlMoveMemory" (Destination As Any, _
                                                                    Source As Any, _
                                                                    ByVal Length As Long)


' ------------------------------------------------------- CONSTANTS
Private Const VT_BY_REF = &H4000&
Private Const FILE_SHARE_READ = &H1
Private Const FILE_SHARE_WRITE = &H2
Private Const MOVEFILE_REPLACE_EXISTING = &H1
Private Const FILE_ATTRIBUTE_TEMPORARY = &H100
Private Const FILE_ATTRIBUTE_NORMAL = &H80
Private Const FILE_BEGIN = 0
Private Const CREATE_NEW = 1
Private Const OPEN_EXISTING = 3
Private Const OPEN_ALWAYS = 4
Private Const GENERIC_READ = &H80000000
Private Const GENERIC_WRITE = &H40000000
Private Const PAGE_READWRITE = 4
Private Const FILE_MAP_WRITE = &H2
Private Const FILE_MAP_READ = &H4
Private Const FADF_FIXEDSIZE = &H10


' ======================================================================================
'  OPEN and CLOSE the MemoryMappedFile
' ======================================================================================

Private FileHandle As Long
Private MappingAddress As Long
Private FileLength As Long

Public Sub OpenMMF(Optional ByVal Filename As String = "Theremino1", _
                   Optional ByVal Length As Long = 4096)

    ' --------------------------------------------------------- MAP THE FILE
    FileHandle = CreateFileMapping(0, _
                                 0, _
                                 PAGE_READWRITE, _
                                 0, _
                                 Length, _
                                 Filename)
    
    If FileHandle = 0 Then CloseMMF: Exit Sub
    
    MappingAddress = MapViewOfFile(FileHandle, _
                             FILE_MAP_WRITE, _
                             0, 0, 0)

    If MappingAddress = 0 Then CloseMMF: Exit Sub
    
    FileLength = Length
End Sub

Public Sub CloseMMF()
    If MappingAddress <> 0 Then UnmapViewOfFile MappingAddress
    If FileHandle <> 0 Then CloseHandle FileHandle
End Sub

' ---------------------------------------------------------------------------- String
Public Sub WriteString(ByVal Offset As Long, ByVal Text As String)
    If MappingAddress = 0 Then Exit Sub
    If Offset < 0 Or Offset >= FileLength Then Exit Sub
    Dim i As Long
    Text = Text & Chr(0)
    For i = 0 To Len(Text) - 1
        CopyMemoryWrite MappingAddress + i + Offset, CByte(Asc(Mid(Text, i + 1, 1))), 1
    Next
End Sub
Public Function ReadString(ByVal Offset As Long) As String
    If MappingAddress = 0 Then ReadString = "": Exit Function
    If Offset < 0 Or Offset >= FileLength Then ReadString = "": Exit Function
    Dim i As Long
    ReadString = ""
    Dim b As Byte
    For i = Offset To Offset + FileLength - 1
        CopyMemoryRead b, MappingAddress + i, 1
        If b = 0 Then Exit For
        ReadString = ReadString & Chr(b)
    Next
End Function

' ---------------------------------------------------------------------------- Long (Int32)
Public Sub WriteInt32(ByVal Offset As Long, ByVal Value As Long)
    If MappingAddress = 0 Then Exit Sub
    If Offset < 0 Or Offset >= FileLength Then Exit Sub
    CopyMemoryWrite MappingAddress + Offset, Value, 4
End Sub
Public Function ReadInt32(ByVal Offset As Long) As Long
    If MappingAddress = 0 Then ReadInt32 = 0: Exit Function
    If Offset < 0 Or Offset >= FileLength Then ReadInt32 = 0: Exit Function
    CopyMemoryRead ReadInt32, MappingAddress + Offset, 4
End Function

' ---------------------------------------------------------------------------- Single
Public Sub WriteSingle(ByVal Offset As Long, ByVal Value As Single)
    If MappingAddress = 0 Then Exit Sub
    If Offset < 0 Or Offset >= FileLength Then Exit Sub
    Dim i As Long
    i = SingleToLong(Value)
    CopyMemoryWrite MappingAddress + Offset, i, 4
End Sub
Public Function ReadSingle(ByVal Offset As Long) As Single
    If MappingAddress = 0 Then ReadSingle = 0: Exit Function
    If Offset < 0 Or Offset >= FileLength Then ReadSingle = 0: Exit Function
    Dim i As Long
    CopyMemoryRead i, MappingAddress + Offset, 4
    ReadSingle = LongToSingle(i)
End Function
Private Function LongToSingle(ByRef i As Long) As Single
    CopyMemory LongToSingle, i, 4
End Function
Private Function SingleToLong(ByRef i As Single) As Long
    CopyMemory SingleToLong, i, 4
End Function


' ======================================================================================
'  WriteSlot and ReadSlot
' ======================================================================================
Public Sub WriteSlot(ByVal Slot As Long, ByVal Value As Single)
    WriteSingle Slot * 4, Value
End Sub
Public Function ReadSlot(ByVal Slot As Long) As Single
    ReadSlot = ReadSingle(Slot * 4)
End Function

